/* graphio.cc
 * Copyright (C) 2001-2 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <pwd.h>
#include "graphio.h"
#include "misc.h"

GraphBlock::GraphBlock()
{
	gap = 0;
	howlong = 0;
	length = 0;
	start = 0;
	average = 0;
	offset = 0;
	time = 0;
	wrap = FALSE;
}

GraphIO::GraphIO()
{
	char *md;
	struct passwd *pwd;

	md = getenv("NOMAD_DIR");
	if (md!=NULL) { strcpy(datadir, md); } else
	{ pwd = getpwuid(getuid()); sprintf(datadir, "%s/.nomad", pwd->pw_dir); }

	replace_flag = FALSE;
}

GraphIO::~GraphIO()
{
}

void GraphIO::init_local()
{
	int i, count;

	num_blocks = num_global_blocks;
	count = GRAPH_FILE_HEADER_SIZE+(GRAPH_BLOCK_HEADER_SIZE*num_blocks);
	for (i=0; i<num_blocks; i++)
	{
		blocks[i].gap     = global_blocks[i].gap;
		blocks[i].howlong = global_blocks[i].howlong;
		blocks[i].length  =
			(((blocks[i].howlong/blocks[i].gap)+1)*record_size);
		blocks[i].start   = count;
		blocks[i].average = count+blocks[i].length-record_size;
		blocks[i].offset  = count;
		blocks[i].time    = 0;
		blocks[i].wrap = FALSE;
		count += blocks[i].length;
	}
}

// compare global and local settings to make sure they are in sync
int GraphIO::valid()
{
	int i;

	if (num_global_blocks!=num_blocks) return 0;
	for (i=0; i<num_blocks; i++)
	{
		if (global_blocks[i].gap     != blocks[i].gap) return 0;
		if (global_blocks[i].howlong != blocks[i].howlong) return 0;
	}
	return 1;
}

void GraphIO::load_header(FILE *f)
{
	int i, j;
	char ch;

	fseek(f, 4L, SEEK_SET);
	num_blocks = read_int32(f);
	for (i=0; i<num_blocks; i++)
	{
		blocks[i].gap     = read_ulong32(f);
		blocks[i].howlong = read_ulong32(f);
		blocks[i].length  = read_ulong32(f);
		blocks[i].start   = read_ulong32(f);
		blocks[i].average = read_ulong32(f);
		blocks[i].offset  = read_ulong32(f);
		blocks[i].time    = read_ulong32(f);
		ch = fgetc(f); 
		blocks[i].wrap = (ch=='y')?TRUE:FALSE;
	}
}

// create a new file using global settings
int GraphIO::create(char *fname)
{
	FILE *f;
	char buf[8], ch;
	int i, j, count;

	init_local();

	f = fopen(fname, "w");
	if (!f) return 0;

	count = 0;
	strcpy(buf, "NMD1"); for (i=0; i<4; i++) fputc(buf[i], f);
	write_int32(f, num_blocks);
	for (i=0; i<num_blocks; i++)
	{
		write_ulong32(f, blocks[i].gap);
		write_ulong32(f, blocks[i].howlong);
		write_ulong32(f, blocks[i].length);
		write_ulong32(f, blocks[i].start);
		write_ulong32(f, blocks[i].average);
		write_ulong32(f, blocks[i].offset);
		write_ulong32(f, blocks[i].time);
		ch = (blocks[i].wrap==TRUE)?'y':'n';
		fputc((int)ch, f);
		count += blocks[i].length;
	}
	for (i=0; i<count; i++) fputc(0, f);
	fclose(f);

	return 1;
}

// add local data to file
int GraphIO::add(char *fname)
{
	FILE *f;

	f = fopen(fname, "r");
	if (!f)
	{
		if (create(fname)==0) return 0;
	}else
	{
		load_header(f);
		if ((!valid()) && (replace_flag))
		{
			fclose(f);
			if (create(fname)==0) return 0;
		}
	}

	f = fopen(fname, "r+");
	if (f)
	{
		add_recursive(f, 0);
		fclose(f);
	}
	return 1;
}

void GraphIO::add_recursive(FILE *f, int lvl)
{
	unsigned long tmpvals[GRAPH_NUM_TOTAL];
	TimeStamp tmptime, oldtime;
	char tmpflag, ch;
	long tsl;
	int i;
	// if removing last entry throw it away
	//
	if (lvl>=num_blocks) return;

	// load current average for block
	//
	fseek(f, blocks[lvl].average, SEEK_SET);
	tmptime.read_time64(f);
	tmpflag = fgetc(f);
	for (i=0; i<num_values; i++)
		tmpvals[i] = read_ulong64(f);

	// now merge in new record to get a new average
	//
	tmptime = current;
	if (status=='1') tmpflag='1';
	for (i=0; i<num_values; i++)
		tmpvals[i] = (unsigned long)(((((double)tmpvals[i])/(blocks[lvl].time+1))*blocks[lvl].time)+((double)value[i])/(blocks[lvl].time+1));

	// get timestamp of previous record
	//
	if (blocks[lvl].offset - record_size < blocks[lvl].start)
		tsl = blocks[lvl].average - record_size;
	else
		tsl = blocks[lvl].offset - record_size;
	fseek(f, tsl, SEEK_SET);
	oldtime.read_time64(f);

	if (tmptime.compare(oldtime) >= blocks[lvl].gap)
	{
		// tmpvals, tmptime, tmpflag are the average values for this level
		if (blocks[lvl].wrap==TRUE)
		{
			// copy out record being erased to current, status, value[]
			// now ready to be passed on to next recursion
			fseek(f, blocks[lvl].offset, SEEK_SET);
			current.read_time64(f);
			status = fgetc(f);
			for (i=0; i<num_values; i++)
				value[i] = read_ulong64(f);
		}

		// write record to blocks[lvl].offset
		fseek(f, blocks[lvl].offset, SEEK_SET);
		tmptime.write_time64(f);
		fputc(tmpflag, f);
		for (i=0; i<num_values; i++)
			write_ulong64(f, tmpvals[i]);

		// reinitialise average
		fseek(f, blocks[lvl].average, SEEK_SET);
		for (i=0; i<record_size; i++) fputc(0, f);
		blocks[lvl].time = 0;

		if (blocks[lvl].wrap==TRUE)
			add_recursive(f, lvl+1);

		// advance offset for block (wrap if needed)
		blocks[lvl].offset += record_size;
		if (blocks[lvl].offset == blocks[lvl].average)
		{
			blocks[lvl].offset = blocks[lvl].start;
			blocks[lvl].wrap = TRUE;
		}
	}else
	{
		// write back the new average values
		fseek(f, blocks[lvl].average, SEEK_SET);
		tmptime.write_time64(f);
		fputc(tmpflag, f);
		for (i=0; i<num_values; i++)
			write_ulong64(f, tmpvals[i]);
		blocks[lvl].time += 1;
	}

	// write back block header
	fseek(f, GRAPH_FILE_HEADER_SIZE+(lvl*GRAPH_BLOCK_HEADER_SIZE)+20, SEEK_SET);
	write_ulong32(f, blocks[lvl].offset);
	write_ulong32(f, blocks[lvl].time);
	ch = (blocks[lvl].wrap==TRUE)?'y':'n';
	fputc((int)ch, f);
}

// get data from file and store in DataSet ready for graph drawing code
gboolean GraphIO::get(char *fname, DataSet *d)
{
	int i, j, len, count;
	int num_in_block;
	FILE *f;

	f = fopen(fname, "r");
	if (!f) return FALSE;

	printf("############################### Grabbing GraphIO::get\n");

	load_header(f);

	printf("Header has %d blocks\n", num_blocks);

	// work out size for data structures
	count = 0;
	for (i=0; i<num_blocks; i++)
	{
		num_in_block = (blocks[i].length / record_size) - 1;
		printf("Block %d - %s gap=%lu howlong=%lu length=%lu start=%lu offset=%lu average=%lu time=%lu num=%d\n", i, ((blocks[i].wrap==TRUE)?"WRAPPED":"NO WRAP"), blocks[i].gap, blocks[i].howlong, blocks[i].length, blocks[i].start, blocks[i].offset, blocks[i].average, blocks[i].time, num_in_block);
		count += num_in_block;
	}
	d->num_entries = count;
	d->num_variables = num_values;

	// free up existing memory allocation
	if (d->timestamps != NULL) delete [] d->timestamps;
	if (d->variables != NULL) delete []  d->variables;
	if (d->status != NULL) delete []  d->status;

	// allocate memory for data structures
	d->timestamps = new TimeStamp[count];
	d->status = new char[count];
	d->variables = new DataLine[num_values];
	for (i=0; i<num_values; i++)
	{
		d->variables[i].values = new unsigned long[count];
	}

	count = 0;
	for (i=num_blocks-1; i>=0; i--)
	{
		if (blocks[i].wrap==TRUE)
		{
			len = ((blocks[i].start+blocks[i].length-blocks[i].offset)/record_size)-1;
			fseek(f, blocks[i].offset, SEEK_SET);
			for (j=0; j<len; j++)
			{
				get_record(d, f, count);
				count++;
			}
		}
		len = (blocks[i].offset-blocks[i].start)/record_size;
		fseek(f, blocks[i].start, SEEK_SET);
		for (j=0; j<len; j++)
		{
			get_record(d, f, count);
printf("#### RECORD #### - ts=%s\n", d->timestamps[count].date_to_str(1));
			count++;
		}
	}
	fclose(f);

	return TRUE;
}

void GraphIO::get_record(DataSet *d, FILE *f, int count)
{
	int j;

	d->timestamps[count].read_time64(f);
	d->status[count] = fgetc(f);
	for (j=0; j<num_values; j++)
		d->variables[j].values[count] = read_ulong64(f);
}

// human readable format for debugging
void GraphIO::print(char *fname)
{
	FILE *f;
	int i, j, k, num_in_block;
	TimeStamp tm;
	char status;
	unsigned long value;

	f = fopen(fname, "r");
	if (f)
	{
		load_header(f);

		printf("============== %s ==============\n", fname);
		printf("Number of blocks = %d\n", num_blocks);
		for (i=0; i<num_blocks; i++)
		{
			num_in_block = (blocks[i].length / record_size);
			printf("Header %d - %s gap=%lu howlong=%lu length=%lu start=%lu offset=%lu average=%lu time=%lu num=%d\n", i, ((blocks[i].wrap==TRUE)?"WRAPPED":"NO WRAP"), blocks[i].gap, blocks[i].howlong, blocks[i].length, blocks[i].start, blocks[i].offset, blocks[i].average, blocks[i].time, num_in_block);
			fseek(f, blocks[i].start, SEEK_SET);
			for (k=0; k<num_in_block; k++)
			{
				tm.read_time64(f);
				status = fgetc(f);
				printf("Timestamp=%18s (%lu) status=%c", tm.date_to_str(1), tm.ts, (status==(char)1)?'1':'0');
				for (j=0; j<num_values; j++)
				{
					value = read_ulong64(f);
					printf(" %lu", value);
				}
				printf("\n");
			}
		}
		fclose(f);
	}
}

